(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Embedding
imports Term

begin

inductive embeds :: "('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> bool" where
  refl: "embeds s s" |
  ctxt: "\<lbrakk> length ss = length ts; \<And> i. i < length ts \<Longrightarrow> embeds (ss ! i) (ts ! i)\<rbrakk> \<Longrightarrow> embeds (Fun f ss) (Fun f ts)" |
  subt: "embeds s t \<Longrightarrow> embeds (Fun f (bef @ s # aft)) t"

lemma embeds_subst: "embeds s t \<Longrightarrow> embeds (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)"
proof (induct rule: embeds.induct)
  case (refl s)
  show "embeds (s \<cdot> \<sigma>) (s \<cdot> \<sigma>)"
    by (rule embeds.refl)
next
  case (ctxt ss ts f)
  show ?case unfolding subst_apply_term.simps 
    by (rule embeds.ctxt, insert ctxt, auto)
next
  case (subt s t f bef aft)
  show ?case unfolding subst_apply_term.simps map_append map.simps
    by (rule embeds.subt, rule subt)
qed

inductive embeds_strict :: "('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> bool" where
  ctxt: "\<lbrakk> length bef = length bef'; \<And> i. i < length bef' \<Longrightarrow> embeds (bef ! i) (bef' ! i);
           length aft = length aft'; \<And> i. i < length aft' \<Longrightarrow> embeds (aft ! i) (aft' ! i);
           embeds_strict s t\<rbrakk> \<Longrightarrow> embeds_strict (Fun f (bef @ s # aft)) (Fun f (bef' @ t # aft'))" |
  subt: "embeds_strict s t \<or> s = t \<Longrightarrow> embeds_strict (Fun f (bef @ s # aft)) t"

lemma embeds_strict_subst: "embeds_strict s t \<Longrightarrow> embeds_strict (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)"
proof (induct rule: embeds_strict.induct)
  case (ctxt bef bef' aft aft' s t f)
  show ?case unfolding subst_apply_term.simps map_append map.simps
    by (rule embeds_strict.ctxt, insert ctxt embeds_subst, force+)
next
  case (subt s t f bef aft)
  thus ?case
  proof
    assume e: "embeds_strict s t \<and> embeds_strict (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)"
    show ?case unfolding subst_apply_term.simps map_append map.simps
      by (rule embeds_strict.subt, rule disjI1, insert e, auto)
  next
    assume e: "s = t"
    show ?case unfolding subst_apply_term.simps map_append map.simps
      by (rule embeds_strict.subt, rule disjI2, unfold e, simp)
  qed
qed

lemma embeds_strict_imp_embeds: "embeds_strict s t \<Longrightarrow> embeds s t"
proof (induct rule: embeds_strict.induct)
  case (ctxt bef bef' aft aft' s t f)
  show ?case 
  proof (rule embeds.ctxt)
    fix i
    assume i: "i < length (bef' @ t # aft')"
    show "embeds ((bef @ s # aft) ! i) ((bef' @ t # aft') ! i)"
    proof (cases "i < length bef'")
      case True
      thus ?thesis using ctxt(1) ctxt(2)[OF True] unfolding nth_append by auto
    next
      case False
      hence "\<exists> j. i = length bef' + j" by arith
      then obtain j where j: "i = length bef' + j" by auto
      with i have i: "j < Suc (length aft')" by auto
      show ?thesis
      proof (cases j)
        case 0
        show ?thesis unfolding j 0 using ctxt unfolding nth_append by auto
      next
        case (Suc jj)
        with i have "jj < length aft'" by auto
        thus ?thesis unfolding j Suc using ctxt unfolding nth_append by auto
      qed
    qed
  qed (insert ctxt, simp)
next
  case (subt s t f bef aft)
  thus ?case
  proof
    assume e: "embeds_strict s t \<and> embeds s t"
    show ?thesis
      by (rule embeds.subt, insert e, auto)
  next
    assume e: "s = t"
    show ?thesis unfolding e
      by (rule embeds.subt, rule embeds.refl)
  qed
qed

lemma embeds_imp_embeds_strict_or_equal: "embeds s t \<Longrightarrow> embeds_strict s t \<or> s = t"
proof (induct rule: embeds.induct)
  case (refl s)
  show ?case by simp
next
  case (ctxt ss ts f)
  show ?case 
  proof (cases "ss = ts")
    case True
    thus ?thesis by simp
  next
    case False
    with nth_equalityI[OF ctxt(1)] obtain i where i': 
      "i < length ss" and neq: "ss ! i \<noteq> ts ! i" by auto
    with ctxt(1) have i: "i < length ts" by auto    
    from ctxt(3)[OF i] neq have siti: "embeds_strict (ss ! i) (ts ! i)" by simp
    let ?ss = "take i ss @ ss ! i # drop (Suc i) ss"
    let ?ts = "take i ts @ ts ! i # drop (Suc i) ts"
    from id_take_nth_drop[OF i'] have ids: "Fun f ss = Fun f ?ss" by simp
    from id_take_nth_drop[OF i] have  idt: "Fun f ts = Fun f ?ts" by simp
    have "embeds_strict (Fun f ss) (Fun f ts)" 
      unfolding ids idt
      by (rule embeds_strict.ctxt, insert siti i ctxt, auto)
    thus ?thesis ..
  qed
next
  case (subt s t f bef aft)
  show ?case
    by (rule disjI1, rule embeds_strict.subt, insert subt, auto)
qed


lemma embeds_size: "embeds s t \<Longrightarrow> size s \<ge> size t"
proof (induct rule: embeds.induct)
  case (refl s)
  show ?case by simp
next
  case (ctxt ss ts f)
  have "list_size size ts \<le> list_size size ss" 
    by (rule list_size_pointwise2, insert ctxt(1) ctxt(3), auto)
  thus ?case by simp
next
  case (subt s t f bef aft)
  thus ?case by auto
qed

lemma embeds_strict_size: "embeds_strict s t \<Longrightarrow> size s > size t"
proof (induct rule: embeds_strict.induct)
  case (subt s t f bef aft)
  thus ?case by auto
next
  case (ctxt bef bef' aft aft' s t f)
  hence len: "length bef' = length bef" "length aft' = length aft" by auto
  note size = embeds_size[OF ctxt(2)] embeds_size[OF ctxt(4)] ctxt(6)
  note size = size[unfolded len]
  have le1: "list_size size bef' \<le> list_size size bef" 
    by (rule list_size_pointwise2, rule ctxt(1)[symmetric], rule size(1))  
  have le2: "list_size size aft' \<le> list_size size aft" 
    by (rule list_size_pointwise2, rule ctxt(3)[symmetric], rule size(2))  
  have "list_size size (bef' @ t # aft') = list_size size bef' + Suc (size t) + list_size size aft'" unfolding list_size_append by simp
  also have "... < list_size size bef + Suc (size s) + list_size size aft"
    using le1 le2 size(3) by auto
  also have "... = list_size size (bef @ s # aft)" by auto
  finally show ?case by simp
qed

lemma embeds_strict_irrefl: "\<not> embeds_strict s s"
proof 
  assume "embeds_strict s s"
  from embeds_strict_size[OF this] show False by simp
qed

lemmas refl = HOL.refl
end