(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Matrix_Poly
imports Linear_Polynomial Matrix Matrix_Complexity
begin

context matrix_carrier 
begin

lemma mat_linear_poly_order_carrier: assumes sd_n: "sd \<le> n" and sd_pos: "sd > 0" shows 
  "linear_poly_order_carrier (mat n n) (mat0 n n) mat_plus (mat_mult n) mat_ge (mat_gt sd) True (mat1 n) (mat_default n) mat_max0 (\<lambda> x. True)"
  by (rule linear_poly_order_carrier.intro, rule linear_poly_carrier.intro,
    rule mat0, 
    rule mat_plus, simp+,
    rule mat_plus_comm,
    rule mat_plus_assoc, simp+,
    simp only: mat_plus_comm[of "mat0 n n"], rule plus_mat0, simp,
    rule mat_mult, simp+,
    rule mat_mult_assoc_n[symmetric], simp+,
    rule mat0_mult_left, simp, 
    rule mat0_mult_right, simp,
    rule mat_mult_plus_distrib_right, simp+,
    rule mat_mult_plus_distrib_left, simp+,
    rule mat_gt_compat[OF sd_n], simp+,
    rule mat_gt_compat2[OF sd_n], simp+,
    rule mat_plus_left_mono, simp+,
    rule mat_mult_left_mono, simp+,
    rule mat_mult_right_mono, simp+,
    rule mat_ge_refl, 
    rule mat_ge_trans, simp+,
    rule mat_gt_trans[OF sd_n], simp+,
    rule mat_plus_gt_left_mono[OF sd_n _ mat_ge_refl], simp+,
    rule linear_poly_order_carrier_axioms.intro,
    simp add:  mat_sn[OF sd_n],
    rule mat_max0_id_pos, simp+,
    rule mat_max0_pos, simp+,
    rule mat_max0_mono, simp+,
    rule mat_max0_x, simp+,
    rule mat_max0, simp,
    rule mat1,
    rule mat1,
    rule mat1_mult_left, simp,
    rule mat1_mult_right, simp,
    rule mat1_ge_mat0,
    rule mat_default_ge_mat0,
    simp_all)
end

  


context mono_matrix_carrier
begin      
lemma mat_mono_linear_poly_order_carrier: assumes sd_n: "sd \<le> n" and sd_pos: "sd > 0" shows 
  "mono_linear_poly_order_carrier (mat n n) (mat0 n n) mat_plus (mat_mult n) mat_ge (mat_gt sd) True (mat1 n) (mat_default n) mat_max0 (\<lambda> x. True) (mat_mono sd) mat_max"
  by (rule mono_linear_poly_order_carrier.intro, rule mat_linear_poly_order_carrier[OF sd_n sd_pos],
    rule mono_linear_poly_order_carrier_axioms.intro,
    simp,
    rule mat_mono[OF sd_n], 
    (simp_all)[6],
    rule mat_default_gt_mat0[OF sd_pos sd_n], 
    simp, 
    (rule mat_max, simp+)[1], 
    (rule mat_max_ge(1), simp+)[1],
    (rule mat_max_ge(2), simp+)[1])

lemma mat_complexity_linear_poly_order_carrier: assumes sd_n: "sd \<le> n" and sd_pos: "sd > 0" shows 
  "complexity_linear_poly_order_carrier (mat n n) (mat0 n n) mat_plus (mat_mult n) mat_ge (mat_gt sd) True (mat1 n) (mat_default n) mat_max0 (\<lambda> x. True) (mat_mono sd) mat_max (\<lambda> a. bound (linear_norm a))"
proof -
  from mat_mono_linear_poly_order_carrier[OF sd_n sd_pos]
  interpret mono_linear_poly_order_carrier "mat n n" "mat0 n n" mat_plus "mat_mult n" mat_ge "mat_gt sd" True "mat1 n" "mat_default n" mat_max0 "\<lambda> x. True" "mat_mono sd" mat_max .
  show ?thesis
  proof(unfold_locales, unfold simp_thms)
    show "bound (linear_norm (mat0 n n)) = 0"
      unfolding linear_norm_0 by (rule bound_zero)
  next    
    fix x y :: "'a mat"    
    assume ge: "mat_ge x y" and x: "mat n n x" and y: "mat n n y"
    show "bound (linear_norm x) \<ge> bound (linear_norm y)"
      by (rule bound_mono, rule linear_norm_ge[OF x y ge])
  next
    fix x y :: "'a mat"    
    assume x: "mat n n x" and y: "mat n n y"
    show "bound (linear_norm (mat_plus x y)) \<le> bound (linear_norm x) + bound (linear_norm y)" unfolding linear_norm_plus[OF x y]
      by (rule bound_plus)
  next
    fix m :: "'a mat"
    let ?m = "mat n n"
    let ?R = "{(x,y). ?m x \<and> ?m y \<and> mat_ge y (mat0 n n) \<and> mat_gt sd x y}"
    let ?f = "\<lambda> m. linear_norm m"
    let ?fR = "{(x,y). y \<succeq> 0 \<and> x \<succ> y}"
    {
      fix x y
      assume "(x,y) \<in> ?R"
      hence x: "?m x" and y: "?m y" and ge: "mat_ge y (mat0 n n)" and gt: "mat_gt sd x y" by auto
      from linear_norm_gt[OF x y gt] linear_norm_ge[OF y mat0 ge, unfolded linear_norm_0]
      have "(?f x, ?f y) \<in> ?fR" by simp
    } note step = this
    show "deriv_bound ?R m (bound (linear_norm m))"
      by (rule deriv_bound_image[of _ "linear_norm", OF bound step])
  qed
qed

lemma linear_poly_order_carrier: "linear_poly_order_carrier (\<lambda> x. True) 0 (op +) (op *) ge gt True 1 d max0 (\<lambda> x. True)"
proof(unfold_locales, simp_all add: field_simps SN max0_x max0_mono max0_pos max0_id_pos plus_left_mono times_left_mono times_right_mono one_ge_zero ge_refl default_ge_zero)
  fix s t u
  assume "s \<succeq> t" and "t \<succ> u"
  thus "s \<succ> u" by (rule compat)
next
  fix s t u
  assume "s \<succ> t" and "t \<succeq> u"
  thus "s \<succ> u" by (rule compat2)
next
  fix s t u
  assume "s \<succeq> t" and "t \<succeq> u"
  thus "s \<succeq> u" by (rule ge_trans)
next
  fix x y z u :: 'a
  assume gt: "x \<succ> y" 
  thus "x + z \<succ> y + z"
    by (rule plus_gt_left_mono)
next
  fix x y z :: 'a
  assume gt: "x \<succ> y" "y \<succ> z"
  thus "x \<succ> z"
    by (rule gt_trans)
qed
end

lemma (in mono_matrix_carrier) mono_linear_poly_order_carrier: "mono_linear_poly_order_carrier (\<lambda> x. True) 0 (op +) (op *) ge gt True 1 d max0 (\<lambda> x. True) mono max"
proof -
  interpret linear_poly_order_carrier "\<lambda> x. True" 0 "op +" "op *" ge gt True 1 d max0 "\<lambda> x. True"
    by (rule linear_poly_order_carrier)
  show ?thesis
    by (unfold_locales, auto simp: mono default_gt_zero)
qed

lemma (in mono_matrix_carrier) complexity_linear_poly_order_carrier: fixes n_of :: "'a \<Rightarrow> nat"
  shows "complexity_linear_poly_order_carrier (\<lambda> x. True) 0 (op +) (op *) ge gt True 1 d max0 (\<lambda> x. True) mono max bound"
proof -
  interpret mono_linear_poly_order_carrier "\<lambda> x. True" 0 "op +" "op *" ge gt True 1 d max0 "\<lambda> x. True" mono max
    by (rule mono_linear_poly_order_carrier)
  show ?thesis
    by (unfold_locales, unfold simp_thms, rule bound, rule bound_plus, rule bound_zero, rule bound_mono)
qed

interpretation int_complexity: mono_matrix_carrier "op > :: int \<Rightarrow> int \<Rightarrow> bool" 1 nat int_mono
proof(unfold_locales, auto)
  fix x
  let ?R = "{(x,y). 0 \<le> (y :: int) \<and> y < x}" 
  show "deriv_bound ?R x (nat x)"
    unfolding deriv_bound_def
  proof
    assume "(\<exists> y. (x,y) \<in> ?R ^^ Suc (nat x))"
    then obtain y where xy: "(x,y) \<in> ?R ^^ Suc (nat x)" ..
    from xy have y: "0 \<le> y" by auto
    obtain n where n: "n = Suc (nat x)" by auto
    from xy[unfolded n[symmetric]]
    have "x \<ge> y + int n"
    proof (induct n arbitrary: x y)
      case 0 thus ?case by auto
    next
      case (Suc n)
      from Suc(2) obtain z where xz: "(x,z) \<in> ?R ^^ n" and zy: "(z,y) \<in> ?R"
        by auto
      from Suc(1)[OF xz] have le: "z + int n \<le> x" .
      from zy have le2: "y + 1 \<le> z" by simp
      with le show ?case by auto
    qed
    with y have nx: "int n \<le> x" by simp
    from nx have x0: "x \<ge> 0" by simp
    with nx n
    show False by simp
  qed      
qed (insert int_SN.mono, auto)

lemma rat_complexity: assumes d0: "d > 0" and d1: "d \<le> 1" and drev: "\<exists> n. d = 1 / of_nat n"
  shows "mono_matrix_carrier (rat_gt d) 1 (\<lambda> x. nat (ceiling (x / d))) rat_mono"
proof -
  let ?gt = "rat_gt d"
  let ?nat = "\<lambda> x. nat (ceiling (x / d))"
  from rat_interpretation[OF d0 d1]
  interpret SN_strict_mono_ordered_semiring_1 1 ?gt rat_mono .
  have id: "(op \<succeq> :: rat \<Rightarrow> rat \<Rightarrow> bool) = (\<lambda> x y. x \<ge> y)" using ge_rat.simps by (intro ext, simp)
  show ?thesis
  proof(unfold_locales, unfold id)
    show "?nat 0 = 0" using d0 by auto
  next
    fix x y :: rat
    assume xy: "x \<ge> y"
    show "?nat x \<ge> ?nat y" 
      by (rule nat_mono, rule ceiling_mono, insert xy d0, auto simp: field_simps)
  next
    fix x y :: rat
    have "ceiling ((x + y) / d) = ceiling (x / d + y / d)" using d0 by (simp add: field_simps)
    also have "... \<le> ceiling (x / d) + ceiling (y / d)" 
      by (rule ceiling_lesseq_add)
    finally show "?nat (x + y) \<le> ?nat x + ?nat y" by auto
  next
    fix x :: rat
    let ?R = "{(x,y). 0 \<le> y \<and> ?gt x y}"
    show "deriv_bound ?R x (?nat x)" unfolding deriv_bound_def
    proof
      assume "(\<exists> y. (x,y) \<in> ?R ^^ Suc (?nat x))"
      then obtain y where xy: "(x,y) \<in> ?R ^^ Suc (?nat x)" ..
      from xy have y: "0 \<le> y" by auto
      obtain n where n: "n = Suc (?nat x)" by auto
      from xy[unfolded n[symmetric]]
      have "x \<ge> y + d * of_nat n"
      proof (induct n arbitrary: x y)
        case 0 thus ?case by auto
      next
        case (Suc n)
        from Suc(2) obtain z where xz: "(x,z) \<in> ?R ^^ n" and zy: "(z,y) \<in> ?R"
          by auto
        from Suc(1)[OF xz] have le: "z + d * of_nat n \<le> x" .
        from zy[unfolded rat_gt_def] have le2: "y + d \<le> z" by simp
        with le show ?case by (auto simp: field_simps)
      qed
      with y have nx: "d * of_nat n \<le> x" by simp
      from d0 have d: "0 \<le> d" by auto
      have "0 \<le> d * of_nat n" by (rule mult_nonneg_nonneg, insert d, auto)
      with nx have x0: "x \<ge> 0" by auto
      have xd0: "0 \<le> x / d"
        by (rule divide_nonneg_pos[OF x0 d0])
      from nx[unfolded n]
      have "d + d * of_nat (?nat x) \<le> x" by (simp add: field_simps)
      with d0 have "d * of_nat (?nat x) < x" by simp 
      also have "of_nat (?nat x) = of_int (ceiling (x / d))" 
        by (rule of_nat_nat, insert xd0, auto)
      also have "of_int (ceiling (x / d)) \<ge> x / d" using ceiling_correct[of "x / d"] by simp
      finally show False using d0 by (auto simp: field_simps)
    qed
  next
    have "1 \<le> nat 1" by simp
    also have "... \<le> ?nat 1"
    proof (rule nat_mono)
      have "1 = ceiling (1 :: rat)" by simp
      also have "... \<le> ceiling (1 / d)"
        by (rule ceiling_mono, insert d1 d0, auto)
      finally show "1 \<le> ceiling (1 / d)" .
    qed
    finally show "1 \<le> ?nat 1" .
  next
    fix x :: rat and n :: nat
    assume x: "0 \<le> x"
    from drev obtain m where d: "d = 1 / of_nat m" by auto
    have "?nat (x + of_nat n) = nat (ceiling (x / d + of_nat n / d))" using d0
      by (simp add: field_simps)
    also have "of_nat n / d = of_int (of_nat (n * m))" unfolding d by (simp add: field_simps of_nat_mult)
    also have "ceiling (x / d + of_int (of_nat (n * m))) = ceiling (x / d) + of_nat (n * m)" unfolding ceiling_add_of_int ..
    also have "nat (ceiling (x / d) + of_nat (n * m)) = ?nat x + nat (int (n * m))" 
    proof (rule int_complexity.bound_plus_of_nat)
      have "x / d \<ge> 0" 
        by (rule divide_nonneg_pos, insert x d0, auto)
      thus "ceiling (x / d) \<succeq> 0" by auto
    qed
    also have "(nat (int (n * m))) = n * m" by simp
    also have "n * m = ?nat (of_nat n)" unfolding d by (simp add: of_nat_mult[symmetric])
    finally
    show "?nat (x + of_nat n) = ?nat x + ?nat (of_nat n)" .
  qed (insert mono, auto)
qed

(* TODO: add times_gt_right_mono to both_mono_ordered_semiring_1 *)
locale both_mono_ordered_semiring_1' = both_mono_ordered_semiring_1 default gt
  for default :: "'a :: ordered_semiring_1"  and gt :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<succ>" 50) + 
  fixes arc_pos :: "'a \<Rightarrow> bool"
  assumes times_gt_right_mono: "gt y z \<Longrightarrow> gt (x * y) (x * z)" 
  and arc_pos_one: "arc_pos (1 :: 'a)"
  and arc_pos_zero: "\<not> arc_pos 0"
  and not_all_ge: "\<And> c d. arc_pos d \<Longrightarrow> \<exists> e. ge e 0 \<and> arc_pos e \<and> \<not> (ge c (d * e))"
begin
lemma scalar_prod_right_mono: assumes wf1: "vec nr v1"
  and wf2: "vec nr v2"
  and wf3: "vec nr v3"
  and gt1: "vec_comp_all gt v2 v3"
  shows "scalar_prod v1 v2 \<succ> scalar_prod v1 v3"
proof -
  from vec_comp_all_index[OF wf2 wf3] gt1 have g1: "\<forall> i < nr. v2 ! i \<succ> v3 ! i" by auto
  from g1 wf1 wf2 wf3 show ?thesis unfolding vec_def
  proof (induct nr arbitrary: v1 v2 v3)
    case (Suc nrr)
    from Suc obtain a1 w1 where v1: "v1 = a1 # w1" and w1: "length w1 = nrr" by (cases v1, auto)
    from Suc obtain a2 w2 where v2: "v2 = a2 # w2" and w2: "length w2 = nrr" by (cases v2, auto)
    from Suc obtain a3 w3 where v3: "v3 = a3 # w3" and w3: "length w3 = nrr" by (cases v3, auto)
    from Suc v2 v3 have a23: "a2 \<succ> a3" and w23: "\<forall> i < nrr. w2 ! i \<succ> w3 ! i" by auto
    have rec: "scalar_prod w1 w2 \<succ> scalar_prod w1 w3" 
      by (rule Suc, auto simp: w1 w2 w3 w23)
    have a: "a1 * a2 \<succ> a1 * a3" by (rule times_gt_right_mono[OF a23])
    show ?case 
      by (simp add: v1 v2 v3 scalar_prod_cons, rule plus_gt_both_mono[OF a rec])
  qed (simp add: scalar_prodI_def zero_leastI)
qed

lemma mat_gt_arc_mult_right_mono: fixes x :: "'a mat" 
  assumes gt1: "mat_gt_arc y z"
  and wfx: "mat nr nc x"
  and wfy: "mat nc ncc y"
  and wfz: "mat nc ncc z"
  shows "mat_gt_arc (mat_mult nr x y) (mat_mult nr x z)"
proof -
  let ?xy = "mat_mult nr x y"
  let ?xz = "mat_mult nr x z"
  from mat_mult[OF wfx wfy] have wfxy: "mat nr ncc ?xy" .
  from mat_mult[OF wfx wfz] have wfxz: "mat nr ncc ?xz" .
  show ?thesis 
  proof (simp only: mat_comp_all_index[OF wfxy wfxz], intro allI impI)
    fix i j
    assume i: "i < ncc" and j: "j < nr"
    have wfxj: "vec nc (row x j)" using row[OF wfx j] .
    have wfyi: "vec nc (col y i)" using col[OF wfy i] .
    have wfzi: "vec nc (col z i)" using col[OF wfz i] .
    show "?xy ! i ! j \<succ> ?xz ! i ! j"
    proof (
        simp only: mat_mult_to_scalar[OF wfx wfy j i],
        simp only: mat_mult_to_scalar[OF wfx wfz j i],
        rule scalar_prod_right_mono[OF wfxj wfyi wfzi],
        simp only: vec_comp_all_index[OF wfyi wfzi],
        intro allI impI
      )
      fix k
      assume k: "k < nc"
      from gt1[unfolded mat_comp_all_index[OF wfy wfz]] i k
      show "col y i ! k \<succ> col z i ! k" unfolding col_def by auto
    qed
  qed
qed

lemma mat_arc_pos_one: assumes n_pos: "n > 0"
  shows "mat_arc_posI arc_pos (mat1 n)"
  unfolding mat_arc_posI_def 
  unfolding mat1_index[OF n_pos n_pos] using arc_pos_one by simp

lemma mat_arc_pos_zero: assumes n_pos: "n > 0"
  shows "\<not> mat_arc_posI arc_pos (mat0 n n)"
  unfolding mat_arc_posI_def 
  unfolding mat0_index[OF n_pos n_pos] using arc_pos_zero by simp
  
end

locale both_mono_matrix_carrier = SN_both_mono_ordered_semiring_1 d gt + both_mono_ordered_semiring_1' d gt
  for gt :: "'a :: {show,ordered_semiring_1} \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<succ>" 50) and d :: 'a
begin

lemma mat_not_all_ge: assumes n_pos: "n > 0"
  and m1: "mat n n m1"
  and m2: "mat n n m2"
  and a2: "mat_arc_pos m2"
  shows "\<exists> m. mat n n m \<and> mat_ge m (mat0 n n) \<and> mat_arc_pos m \<and> \<not> mat_ge m1 (mat_mult n m2 m)"
proof -
  obtain c where c: "c = m1 ! 0 ! 0" by auto
  from a2 have "arc_pos (m2 ! 0 ! 0)" unfolding mat_arc_posI_def .
  from not_all_ge[OF this, of c] obtain e where e0: "e \<succeq> 0" and ae: "arc_pos e"
    and nc: "\<not> c \<succeq> m2 ! 0 ! 0 * e" by auto
  let ?gen = "\<lambda> f. map (\<lambda>i. map (f i) [0..<n]) [0..<n]"
  {
    fix f :: "nat \<Rightarrow> nat \<Rightarrow> 'a" 
    have "mat n n (?gen f)"
      unfolding mat_def list_all_iff by (auto simp: vec_def)    
  } note gen = this
  obtain f :: "nat \<Rightarrow> nat \<Rightarrow> 'a" where f: "f = (\<lambda> i j. if i = 0 \<and> j = 0 then e else 0)" by auto
  let ?m = "?gen f"
  have m00: "?m ! 0 ! 0 = e" using n_pos unfolding f by auto
  show ?thesis
  proof (rule exI[of _ ?m], intro conjI, rule gen)
    show "mat_ge ?m (mat0 n n)"
      unfolding mat_ge_index[OF gen mat0]
    proof (intro allI impI)
      fix i j
      assume i: "i < n" and j: "j < n"
      have m: "?m ! i ! j = f i j" using i j by auto
      have 0: "mat0 n n ! i ! j = (0 :: 'a)" using mat0_index[OF i j] .
      show "?m ! i ! j \<succeq> mat0 n n ! i ! j" unfolding m 0
        unfolding f using e0 ge_refl by auto
    qed
  next
    show "mat_arc_pos ?m" 
      unfolding mat_arc_posI_def 
      unfolding m00 by (rule ae)
  next
    let ?mult = "mat_mult n m2 ?m"
    from n_pos obtain nn where n: "n = Suc nn" by (cases n, auto)
    have col: "col ?m 0 = map (f 0) [0 ..< n]" unfolding col_def using n_pos by simp
    also have "... = f 0 0 # map (\<lambda> i. f 0 (Suc i)) [0 ..< nn]"
      unfolding n unfolding map_nth_Suc ..
    also have "... = e # replicate nn 0" unfolding f 
      by (simp add: map_replicate_trivial)
    also have "... = e # vec0 nn" unfolding vec0I_def ..
    finally have col: "col ?m 0 = e # vec0 nn" by simp
    from row[OF m2 n_pos] have row: "length (row m2 0) = n" unfolding vec_def .
    with n_pos obtain r where row: "row m2 0 = (row m2 0 ! 0) # r"      
      by (cases "row m2 0", auto)
    also have "... = m2 ! 0 ! 0 # r" unfolding row_col[OF m2 n_pos n_pos]
      unfolding col_def ..
    finally have row: "row m2 0 = m2 ! 0 ! 0 # r" by simp
    from mat_mult_to_scalar[OF m2 gen n_pos n_pos]
    have "?mult ! 0 ! 0 = scalar_prod (row m2 0) (col ?m 0)"
      by simp
    also have "... = scalar_prod (m2 ! 0 ! 0 # r) (e # vec0 nn)"
      unfolding row col ..
    also have "... = m2 ! 0 ! 0 * e + scalar_prod r (vec0 nn)" 
      unfolding scalar_prod_cons ..
    also have "... = m2 ! 0 ! 0 * e"
      unfolding scalar_right_zero by simp
    finally have "?mult ! 0 ! 0 = m2 ! 0 ! 0 * e" .
    with nc c have "\<not> m1 ! 0 ! 0 \<succeq> ?mult ! 0 ! 0" by simp
    thus "\<not> mat_ge m1 ?mult"
      unfolding mat_ge_index[OF m1 mat_mult[OF m2 gen]] using n_pos
      by auto
  qed
qed


lemma mat_linear_poly_order_carrier: assumes n_pos: "n > 0" shows 
  "linear_poly_order_carrier (mat n n) (mat0 n n) mat_plus (mat_mult n) mat_ge mat_gt_arc False (mat1 n) (mat_default n) id (mat_arc_posI arc_pos)"
  by (rule linear_poly_order_carrier.intro, rule linear_poly_carrier.intro, 
    rule mat0, 
    rule mat_plus, simp+,
    rule mat_plus_comm,
    rule mat_plus_assoc, simp+,
    simp only: mat_plus_comm[of "mat0 n n"], rule plus_mat0, simp,
    rule mat_mult, simp+,
    rule mat_mult_assoc_n[symmetric], simp+,
    rule mat0_mult_left, simp, 
    rule mat0_mult_right, simp,
    rule mat_mult_plus_distrib_right, simp+,
    rule mat_mult_plus_distrib_left, simp+,
    rule mat_gt_arc_compat, simp+,
    rule mat_gt_arc_compat2, simp+,
    rule mat_plus_left_mono, simp+,
    rule mat_mult_left_mono, simp+,
    rule mat_mult_right_mono, simp+,
    rule mat_ge_refl, 
    rule mat_ge_trans, simp+, 
    rule mat_gt_arc_trans, simp+,
    rule mat_gt_arc_plus_mono, simp+,
    rule mat_gt_arc_mult_mono, simp+,
    rule mat_gt_arc_mult_right_mono, simp+,
    rule mat0_leastI, simp+,
    rule mat0_leastII, simp+,
    rule linear_poly_order_carrier_axioms.intro,
    rule SN_subset[OF mat_gt_arc_SN[OF n_pos]], force,
    simp,
    simp add: mat0_leastIII,
    simp,
    simp add: mat_ge_refl,
    simp,
    rule mat1,
    rule mat1, 
    rule mat1_mult_left, simp,
    rule mat1_mult_right, simp,
    rule mat1_ge_mat0,
    rule mat_default_ge_mat0,
    rule mat_arc_pos_plus[OF n_pos], simp+,
    rule mat_arc_pos_mult[OF n_pos], simp+,
    rule mat_arc_pos_mat_default[OF n_pos],
    simp,
    rule mat_arc_pos_one[OF n_pos],
    rule mat_arc_pos_zero[OF n_pos],
    rule mat_not_all_ge[OF n_pos]) 
end

lemma (in both_mono_matrix_carrier) linear_poly_order_carrier: "linear_poly_order_carrier (\<lambda> x. True) 0 (op +) (op *) ge gt False 1 d id arc_pos"
proof(unfold_locales)
  fix s t u
  assume "s \<succ> t" and "t \<succeq> u"
  thus "s \<succ> u" by (rule compat2)
next
  fix s t u
  assume "s \<succeq> t" and "t \<succeq> u"
  thus "s \<succeq> u" by (rule ge_trans)
next
  fix s t u
  assume "s \<succ> t" and "t \<succ> u"
  thus "s \<succ> u" by (rule gt_trans)
qed (insert not_all_ge, simp_all add: field_simps plus_left_mono times_left_mono times_right_mono zero_leastI SN zero_leastII zero_leastIII one_ge_zero ge_refl arc_pos_default arc_pos_plus arc_pos_mult compat plus_gt_both_mono times_gt_left_mono times_gt_right_mono arc_pos_one arc_pos_zero)
  
 
end
