(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Outermost_Rewriting
imports
  Trs
  Context_Substitution
  "~~/src/HOL/Library/Monad_Syntax"
begin

definition ostep_cond_single :: "('f,'v)term \<Rightarrow> pos \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
  where "ostep_cond_single l p t \<equiv> \<forall> q < p. \<not> (\<exists> \<sigma>. t |_ q = l \<cdot> \<sigma>)"

definition ostep_cond :: "('f,'v)term set \<Rightarrow> pos \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
  where "ostep_cond Q p t \<equiv> \<forall> l \<in> Q. ostep_cond_single l p t"

definition ostep_p :: "('f,'v)term set \<Rightarrow> ('f,'v)trs \<Rightarrow> pos \<Rightarrow> ('f,'v)trs" 
  where "ostep_p Q R p \<equiv> {(s,t). (\<exists> lr \<sigma>. (s,t) \<in> rstep_r_p_s R lr p \<sigma>) \<and> ostep_cond Q p s}"

definition ostep :: "('f,'v)term set \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs"
  where "ostep Q R \<equiv> {(s,t). \<exists> p. (s,t) \<in> ostep_p Q R p}"

lemmas ostep_defs = ostep_def ostep_p_def

lemma ostep_pI: "(s,t) \<in> rstep_r_p_s R lr p \<sigma> \<Longrightarrow> ostep_cond Q p s \<Longrightarrow> (s,t) \<in> ostep_p Q R p" 
  unfolding ostep_defs by blast

lemma ostep_pE[elim]: "(s,t) \<in> ostep_p Q R p \<Longrightarrow> \<lbrakk>\<And> \<sigma> lr. (s,t) \<in> rstep_r_p_s R lr p \<sigma> \<Longrightarrow> ostep_cond Q p s \<Longrightarrow> P\<rbrakk> \<Longrightarrow> P"
  unfolding ostep_defs by blast

type_synonym ('f,'v)oloop = "('f,'v)ctxt \<times> ('f,'v)subst \<times> nat \<times> (nat \<Rightarrow> ('f,'v)term) \<times> (nat \<Rightarrow> pos)"

definition oloop :: "('f,'v)term set \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)oloop \<Rightarrow> bool"
  where "oloop Q R ol \<equiv> case ol of (C,\<mu>,len,ts,ps) \<Rightarrow>
  (\<forall> i \<le> len. (\<exists> \<sigma> lr. (ts i, ts (Suc i)) \<in> rstep_r_p_s R lr (ps i) \<sigma>)) \<and>
  (\<forall> i \<le> len. \<forall> n. ostep_cond Q (hole_pos C ^ n <#> ps i) (ctxt_subst C \<mu> n (ts i))) \<and>
  (ts (Suc len) = C\<langle>ts 0 \<cdot> \<mu>\<rangle>)"

lemma oloop_cond: assumes loop: "oloop Q R (C,\<mu>,len,ts,ps)"
  and i: "i \<le> len"
  shows "ostep_cond Q (hole_pos C ^ n <#> ps i) (ctxt_subst C \<mu> n (ts i))"
  using assms unfolding oloop_def by blast

lemma oloop_ostep_p: assumes oloop: "oloop Q R (C,\<mu>,len,ts,ps)"
  and i: "i \<le> len"
  shows "(ctxt_subst C \<mu> n (ts i), ctxt_subst C \<mu> n (ts (Suc i))) \<in> ostep_p Q R (hole_pos C ^ n <#> ps i)"
proof -
  note loop = oloop[unfolded oloop_def]
  from loop i obtain \<sigma> lr where step: "(ts i, ts (Suc i)) \<in> rstep_r_p_s R lr (ps i) \<sigma>" by auto
  show ?thesis
    by (rule ostep_pI[OF ctxt_subst_step[OF step]], rule oloop_cond[OF oloop], insert i, auto)
qed

lemma oloop_end: assumes oloop: "oloop Q R (C,\<mu>,len,ts,ps)"
  shows "ctxt_subst C \<mu> n (ts (Suc len)) = ctxt_subst C \<mu> (Suc n) (ts 0)"
proof -
  note loop = oloop[unfolded oloop_def]
  from loop have ts: "ts (Suc len) = C\<langle>ts 0 \<cdot> \<mu>\<rangle>" by auto
  show ?thesis unfolding ts unfolding ctxt_subst_Suc ..
qed

lemma oloop_iterate: assumes oloop: "oloop Q R (C,\<mu>,len,ts,ps)"
  shows "(ctxt_subst C \<mu> n (ts 0), ctxt_subst C \<mu> (Suc n) (ts 0)) \<in> (ostep Q R)^+"
proof -
  let ?t = "\<lambda> i. ctxt_subst C \<mu> n (ts i)"
  from oloop_ostep_p[OF oloop] have steps_i: "\<And> i. i \<le> len \<Longrightarrow> (?t i, ?t (Suc i)) \<in> ostep Q R" unfolding ostep_def by blast
  have steps: "(?t 0, ?t len) \<in> (ostep Q R)^*"
    unfolding rtrancl_fun_conv
    by (rule exI[of _ ?t], rule exI[of _ len], insert steps_i, auto)
  have step: "(?t len, ctxt_subst C \<mu> (Suc n) (ts 0)) \<in> ostep Q R"
    using steps_i[of len]
    unfolding oloop_end[OF oloop, of n] by simp
  from steps step show ?thesis by auto
qed

lemma oloop_imp_not_SN: assumes oloop: "oloop Q R (C,\<mu>,len,ts,ps)"
  shows "\<not> SN_on (ostep Q R) {ts 0}"
proof
  assume SN: "SN_on (ostep Q R) {ts 0}"
  let ?O = "(ostep Q R)^+"
  from SN have SN: "SN_on ?O {ts 0}" by (rule SN_on_trancl)
  let ?t = "\<lambda> i. ctxt_subst C \<mu> i (ts 0)"
  from oloop_iterate[OF oloop] have steps: "\<And> i. (?t i, ?t (Suc i)) \<in> ?O" .
  have steps: "\<exists> f. f 0 \<in> {ts 0} \<and> (\<forall> i. (f i, f (Suc i)) \<in> ?O)" 
    by (rule exI[of _ ?t], insert steps, auto)
  with SN show False unfolding SN_defs by blast
qed

definition ostep_ctxt_subst_cond :: "('f,'v)term \<Rightarrow> ('f,'v)ctxt \<Rightarrow> ('f,'v)subst \<Rightarrow> pos \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
  where "ostep_ctxt_subst_cond l C \<mu> p t \<equiv> (\<forall> n. ostep_cond_single l (hole_pos C ^ n <#> p) (ctxt_subst C \<mu> n t))"

lemma oloopI: assumes steps: "\<And> i. i \<le> len \<Longrightarrow> (ts i, ts (Suc i)) \<in> rstep_r_p_s R (lr i) (ps i) (\<sigma> i)"
  and len: "ts (Suc len) = C\<langle>ts 0 \<cdot> \<mu>\<rangle>"
  and ocond: "\<And> l i. i \<le> len \<Longrightarrow> ps i \<in> poss (ts i) \<Longrightarrow> l \<in> Q \<Longrightarrow> ostep_ctxt_subst_cond l C \<mu> (ps i) (ts i)"
  shows "oloop Q R (C,\<mu>,len,ts,ps)"
  unfolding oloop_def split_def fst_conv snd_conv
proof (intro conjI allI impI)
  fix i
  assume "i \<le> len"
  from steps[OF this]
  show "\<exists> \<sigma> lr. (ts i, ts (Suc i)) \<in> rstep_r_p_s R lr (ps i) \<sigma>" by blast
next
  show "ts (Suc len) = C\<langle>ts 0 \<cdot> \<mu>\<rangle>" by (rule len)
next
  fix i n
  assume i: "i \<le> len"
  from steps[OF i] have "ps i \<in> poss (ts i)" unfolding rstep_r_p_s_def' by auto
  from ocond[OF i this, unfolded ostep_ctxt_subst_cond_def]
  show "ostep_cond Q (hole_pos C ^ n <#> ps i) (ctxt_subst C \<mu> n (ts i))" 
    unfolding ostep_cond_def by blast
qed

end


