theory Termination_Inheritance
imports
  Relative_Rewriting
  Regexp_Method
  Seq
begin

lemma 1: "r^+ \<subseteq> (relto r s)^+" by regexp
lemma 2: "(relto r s)^+ \<subseteq> (r \<union> s)^+" by regexp
lemma 3: "(relto r s)^+ = (relto r (r \<union> s))^+" by regexp
lemma 4: "(relto r s)^+ = (relto r (s - r))^+"
  unfolding 3[of "s - r" r] by (simp) regexp

lemma 5: "(relto (r \<union> s) q)^+ = (relto r (s \<union> q))^+ \<union> (relto s q)^+"
  by regexp

lemma 6:
  "(relto (r \<union> s) (q1 \<union> q2))^+ \<subseteq> (relto (r \<union> q1) (s \<union> q2))^+ \<union> (relto s q2)^+"
  by regexp


text {*Termination inheritance by transitivity (see, e.g., Geser's thesis).*}

lemma trans_subset_SN:
  assumes "trans R" and "R \<subseteq> (r \<union> s)" and "SN r" and "SN s"
  shows "SN R"
proof
  fix f :: "nat \<Rightarrow> 'a"
  assume "f 0 \<in> UNIV"
    and seq: "\<forall>i. (f i, f (Suc i)) \<in> R"
  have trans_seq: "\<forall>i j. i < j \<longrightarrow> (f i, f j) \<in> r \<union> s"
    using assms and seq_imp_trancl[OF seq] by auto
  let ?M = "{i::nat. \<forall>j>i. (f i, f j) \<notin> r}"
  show False
  proof (cases "finite ?M")
    case True
    let ?n = "Max ?M"
    from Max_ge[OF True] have "\<forall>i\<in>?M. i \<le> ?n" by simp
    hence "\<forall>k\<ge>Suc ?n. \<exists>k'>k. (f k, f k') \<in> r" by auto
    from steps_imp_seq[of "Suc ?n" "\<lambda>x y. (x, y) \<in> r", OF this] and assms
      show False by auto
  next
    case False
    from choice[OF this[unfolded infinite_nat_iff_unbounded]]
      obtain g where 1: "\<forall>i. g i > i \<and> g i \<in> ?M" by auto
    with trans_seq have 2: "\<forall>i\<ge>0. g i > i \<and> (f i, f (g i)) \<in> r \<union> s" by auto
    let ?g = "\<lambda>i. (g ^^ Suc i) (Suc 0)"
    let ?f = "\<lambda>i. f (?g i)"
    have "\<forall>i. (?f i, ?f (Suc i)) \<in> s"
    proof
      fix i
      from 2 have "(g ^^ i) (Suc n) \<ge> 0" by (induct i) auto
      hence "?g i \<ge> 0" using assms by auto
      with 2[THEN spec[of _ "(g ^^ Suc i) (Suc 0)"]]
        show "(?f i, ?f (Suc i)) \<in> s" using 1 by auto
    qed
    with assms show False by best
  qed
qed

lemma SN_Un_conv:
  assumes "trans (r \<union> s)"
  shows "SN (r \<union> s) \<longleftrightarrow> SN r \<and> SN s"
    (is "SN ?r \<longleftrightarrow> ?rhs")
proof
  assume "SN (r \<union> s)" thus "SN r \<and> SN s"
    using SN_subset[of ?r] by blast
next
  assume "SN r \<and> SN s"
  with trans_subset_SN[OF assms subset_refl] show "SN ?r" by simp
qed

lemma SN_relto_Un:
  "SN (relto (R \<union> S) Q) \<longleftrightarrow> SN (relto R (S \<union> Q)) \<and> SN (relto S Q)"
    (is "SN ?a \<longleftrightarrow> SN ?b \<and> SN ?c")
proof -
  have eq: "?a^+ = ?b^+ \<union> ?c^+" by regexp
  from SN_Un_conv[of "?b^+" "?c^+", unfolded eq[symmetric]]
    show ?thesis unfolding SN_trancl_SN_conv by simp
qed

lemma SN_relto_split:
  assumes "SN (relto r (s \<union> q2) \<union> relto q1 (s \<union> q2))" (is "SN ?a")
    and "SN (relto s q2)" (is "SN ?b")
  shows "SN (relto r (q1 \<union> q2) \<union> relto s (q1 \<union> q2))" (is "SN ?c")
proof -
  have "?c^+ \<subseteq> ?a^+ \<union> ?b^+" by regexp
  from trans_subset_SN[OF _ this, unfolded SN_trancl_SN_conv, OF _ assms]
    show ?thesis by simp
qed

end