(*
Copyright 2012 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Termination_Switch
imports
  Nontermination
  Innermost_Switch 
begin

lemma five_chain_into_three_chain: 
  assumes "ichain (P,{},Q,{},R) s t sigma"
  shows "Nontermination.ichain (P,Q,R) s t sigma"
proof -
  note nfsub = NF_subterm[OF _ supt_imp_supteq, of _ "Id_on Q"]
  from nfsub have a: "\<forall>i. s i \<cdot> sigma i \<in> NF_terms Q \<Longrightarrow> \<forall>i u. s i \<cdot> sigma i \<rhd> u \<longrightarrow> u \<in> NF_terms Q" by blast
  from assms a show ?thesis unfolding QDP_Framework.ichain.simps Nontermination.ichain.simps by force
qed

lemma three_chain_into_five_chain: "Nontermination.ichain (P,{},R) s t sigma \<Longrightarrow> ichain (P,{},{},{},R) s t sigma" 
  by (simp add: ichain.simps)

lemma termination_switch_proc:
  assumes inf: "infinite_dpp (P,{},R)"
  and crit: "critical_pairs P R = {}"
  and WCR: "WCR (rstep R)"
  and NF: "NF_trs R \<subseteq> NF_terms Q"
  and SN: "SN (qrstep Q R) \<Longrightarrow> SN (rstep R)"
  shows "infinite_dpp (P,Q,R)"
proof (cases "SN (qrstep Q R)")
  case False
  thus ?thesis by simp
next
  case True
  from True SN have SNR: "SN (rstep R)" by simp
  thm infinite_dpp.simps
  from inf SNR have nchain: "\<exists> s t sigma. Nontermination.ichain(P,{},R) s t sigma" unfolding infinite_dpp.simps by force
  then obtain s t sigma where nchain: "Nontermination.ichain(P,{},R) s t sigma" by blast
  hence nchain: "ichain (P,{},{},{},R) s t sigma" by (rule three_chain_into_five_chain)
  with SNR have chain: "min_ichain (P,{},{},{},R) s t sigma" 
    by (simp add: minimal_cond_def SN_def)
  from crit have crit: "critical_pairs (P \<union> {}) R = {}" by auto
  have NFempty: "NF_trs R \<subseteq> NF_terms {}" by auto
  from WCR have WCR: "WCR_on (qrstep {} R) {t. SN_on (qrstep {} R) {t}}" 
    unfolding WCR_on_def by auto
  note inn_switch = switch_to_innermost_proc[OF WCR NFempty crit]
  from chain inn_switch have "\<exists> s t sigma. min_ichain (P,{},lhss R,{},R) s t sigma"
    unfolding finite_dpp_def by auto 
  then obtain s t sigma where ichain: "QDP_Framework.ichain (P,{},lhss R,{},R) s t sigma" by auto
  have ichain : "ichain (P,{},Q,{},R) s t sigma"
  proof (rule ichain_mono[OF ichain])
    show "NF_terms (lhss R) \<subseteq> NF_terms Q" using NF by auto
  qed auto
  hence ichain: "Nontermination.ichain (P,Q,R) s t sigma" by (rule five_chain_into_three_chain)
  thus ?thesis unfolding infinite_dpp.simps by blast
qed

end    