(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Util2
  imports Util Archimedean_Field
begin

definition replace :: "'a \<Rightarrow> 'a set \<Rightarrow> 'a set \<Rightarrow> 'a set"
  where "replace a bs M \<equiv> if a \<in> M then M - {a} \<union> bs else M"

definition replace_impl :: "'a \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list"
  where "replace_impl a bs M \<equiv> if a \<in> set M then bs @ filter (\<lambda> b. b \<noteq> a) M else M"

lemma replace_impl[simp]: "set (replace_impl a bs M) = replace a (set bs) (set M)"
  unfolding replace_impl_def replace_def by auto


(* TODO: this function is currently also available in RBT_Map_Set_Extension in AFP/Transitive_Closure *)
(* however, the import of the Collection Framework via RBT_Map_Set_Extension has some undesired 
   behaviour elsewhere
   \<Rightarrow> for next AFP update, move sorted_list_subset out of Transitive_Closure into Util ! *)
fun sorted_list_subset :: "'a :: linorder list \<Rightarrow> 'a list \<Rightarrow> 'a option"
  where "sorted_list_subset (a # as) (b # bs) = 
            (if a = b then sorted_list_subset as (b # bs)
        else if a > b then sorted_list_subset (a # as) bs
        else Some a)"
    |  "sorted_list_subset [] _ = None"
    |  "sorted_list_subset (a # _) [] = Some a"
   
lemma sorted_list_subset: assumes "sorted as" and "sorted bs"
  shows "(sorted_list_subset as bs = None) = (set as \<subseteq> set bs)"
  using assms 
proof (induct rule: sorted_list_subset.induct)
  case (2 bs)
  thus ?case by auto
next
  case (3 a as)
  thus ?case by auto
next
  case (1 a as b bs)
  from 1(3) have sas: "sorted as" and a: "\<And> a'. a' \<in> set as \<Longrightarrow> a \<le> a'" unfolding linorder_class.sorted.simps[of "a # as"] by auto
  from 1(4) have sbs: "sorted bs" and b: "\<And> b'. b' \<in> set bs \<Longrightarrow> b \<le> b'" unfolding linorder_class.sorted.simps[of "b # bs"] by auto
  show ?case
  proof (cases "a = b")
    case True
    from 1(1)[OF this sas 1(4)] True show ?thesis by auto
  next
    case False note oFalse = this
    show ?thesis 
    proof (cases "a > b")
      case True
      with a b have "b \<notin> set as" by force
      with 1(2)[OF False True 1(3) sbs] False True show ?thesis by auto
    next
      case False
      with oFalse have "a < b" by auto
      with a b have "a \<notin> set bs" by force
      with oFalse False show ?thesis by auto
    qed
  qed
qed

text {* functions for removing the i-th element of a list and adjusting an index for this removal *}
definition remove_idx :: "nat \<Rightarrow> 'a list \<Rightarrow> 'a list"
  where "remove_idx i xs \<equiv> take i xs @ drop (Suc i) xs"

lemma remove_idx_len: assumes i: "i < length xs"
  shows "length xs = Suc (length (remove_idx i xs))"
proof -
  show ?thesis unfolding arg_cong[where f = "length", OF id_take_nth_drop[OF i]]
    unfolding remove_idx_def by simp
qed

definition adjust_idx :: "nat \<Rightarrow> nat \<Rightarrow> nat"
  where "adjust_idx i j \<equiv> (if j < i then j else (Suc j))"

definition adjust_idx_rev :: "nat \<Rightarrow> nat \<Rightarrow> nat"
  where "adjust_idx_rev i j \<equiv> (if j < i then j else j - Suc 0)"

lemma adjust_idx_rev1: "adjust_idx_rev i (adjust_idx i j) = j"
  unfolding adjust_idx_def adjust_idx_rev_def by (cases "i < j", auto)

lemma adjust_idx_rev2: assumes "j \<noteq> i" shows "adjust_idx i (adjust_idx_rev i j) = j"
  unfolding adjust_idx_def adjust_idx_rev_def using assms by (cases "i < j", auto)

lemma adjust_idx_i: "adjust_idx i j \<noteq> i" unfolding adjust_idx_def by (cases "j < i", auto)

lemma adjust_idx_nth: assumes i: "i < length xs"
  shows "remove_idx i xs ! j = xs ! adjust_idx i j" (is "?l = ?r")
proof -
  let ?j = "adjust_idx i j"
  from i have ltake: "length (take i xs) = i" by simp
  note nth_xs = arg_cong[where f = "\<lambda> xs. xs ! ?j", OF id_take_nth_drop[OF i], unfolded nth_append ltake]
  show ?thesis
  proof (cases "j < i")    
    case True
    hence j: "?j = j" unfolding adjust_idx_def by simp
    show ?thesis unfolding nth_xs unfolding j remove_idx_def nth_append ltake
      using True by simp
  next
    case False    
    hence j: "?j = Suc j" unfolding adjust_idx_def by simp
    from i have lxs: "min (length xs) i = i" by simp
    show ?thesis unfolding nth_xs unfolding j remove_idx_def nth_append
      using False by (simp add: lxs)
  qed
qed

lemma adjust_idx_rev_nth: assumes i: "i < length xs"
  and ji: "j \<noteq> i"
  shows "remove_idx i xs ! adjust_idx_rev i j = xs ! j" (is "?l = ?r")
proof -
  let ?j = "adjust_idx_rev i j"
  from i have ltake: "length (take i xs) = i" by simp
  note nth_xs = arg_cong[where f = "\<lambda> xs. xs ! j", OF id_take_nth_drop[OF i], unfolded nth_append ltake]
  show ?thesis
  proof (cases "j < i")    
    case True
    hence j: "?j = j" unfolding adjust_idx_rev_def by simp
    show ?thesis unfolding nth_xs unfolding j remove_idx_def nth_append ltake
      using True by simp
  next
    case False
    with ji have ji: "j > i" by auto
    hence j: "?j = j - Suc 0" unfolding adjust_idx_rev_def by simp
    show ?thesis unfolding nth_xs unfolding j remove_idx_def nth_append ltake
      using ji by auto
  qed
qed

lemma adjust_idx_length: assumes i: "i < length xs"
  and j: "j < length (remove_idx i xs)"
  shows "adjust_idx i j < length xs"
  using j unfolding remove_idx_len[OF i] adjust_idx_def by (cases "j < i", auto)

lemma adjust_idx_rev_length: assumes i: "i < length xs"
  and j: "j < length xs"
  and ji: "j \<noteq> i"
  shows "adjust_idx_rev i j < length (remove_idx i xs)"
  using j i ji unfolding remove_idx_len[OF i] adjust_idx_rev_def by (cases "j < i", auto)

lemma [code_unfold]: "(set xs = set ys) = (set xs \<subseteq> set ys \<and> set ys \<subseteq> set xs)"
  by auto

lemma zip_size_aux: "list_size (size o snd) (zip ts ls) \<le> (list_size size ls)"
proof (induct ls arbitrary: ts)
  case Nil thus ?case by auto
next
  case (Cons l ls ts)
  thus ?case by (cases ts, auto)
qed

lemma option_map_cong[fundef_cong]: "xs = ys \<Longrightarrow> \<lbrakk>\<And> x. ys = Some x \<Longrightarrow> f x = g x\<rbrakk> \<Longrightarrow> Option.map f xs = Option.map g ys"
  by (cases ys, auto)

lemma finite_distinct: "finite { xs . distinct xs \<and> set xs = X}" (is "finite (?S X)")
proof (cases "finite X")
  case False
  with finite_set have id: "?S X = {}" by auto
  show ?thesis unfolding id by auto
next
  case True
  show ?thesis
  proof (induct rule: finite_induct[OF True])
    case 1
    show ?case by simp
  next
    case (2 x X)
    let ?L = "{0..< card (insert x X)} \<times> ?S X" 
    from 2(3) 
    have fin: "finite ?L" by auto 
    let ?f = "\<lambda> (i,xs). take i xs @ x # drop i xs"
    show ?case
    proof (rule finite_surj[OF fin, of _ ?f], rule)
      fix xs
      assume "xs \<in> ?S (insert x X)"
      hence dis: "distinct xs" and set: "set xs = insert x X" by auto
      from distinct_card[OF dis] have len: "length xs = card (set xs)" by auto
      from set[unfolded set_conv_nth] obtain i where x: "x = xs ! i" and i: "i < length xs" by auto
      from i have min: "min (length xs) i = i" by simp
      let ?ys = "take i xs @ drop (Suc i) xs"
      from id_take_nth_drop[OF i] have xsi: "xs = take i xs @ xs ! i # drop (Suc i) xs" .
      also have "... = ?f (i,?ys)" unfolding split
        by (simp add: min x)
      finally have xs: "xs = ?f (i,?ys)" .
      show "xs \<in> ?f ` ?L"
      proof (rule image_eqI, rule xs, rule SigmaI)
        show "i \<in> {0..<card (insert x X)}" using i[unfolded len] set[symmetric] by simp
      next
        from dis xsi have disxsi: "distinct (take i xs @ xs ! i # drop (Suc i) xs)" by simp
        note disxsi = disxsi[unfolded distinct_append x[symmetric]]
        have xys: "x \<notin> set ?ys" using disxsi by auto
        from distinct_take_drop[OF dis i]
        have disys: "distinct ?ys" .
        have "insert x (set ?ys) = set xs" unfolding arg_cong[OF xsi, of set] x by simp        
        hence "insert x (set ?ys) = insert x X" unfolding set by simp
        from this[unfolded insert_eq_iff[OF xys 2(2)]]
        show "?ys \<in> ?S X" using disys by auto
      qed
    qed
  qed
qed

lemma finite_distinct_subset: assumes "finite X"
  shows "finite { xs . distinct xs \<and> set xs \<subseteq> X}" (is "finite (?S X)")
proof -
  let ?X = "{ { xs. distinct xs \<and> set xs = Y} | Y. Y \<subseteq> X}"
  have id: "?S X = \<Union> ?X" by blast
  show ?thesis unfolding id
  proof (rule finite_Union)
    show "finite ?X" using assms  by auto
  next
    fix M
    assume "M \<in> ?X"
    with finite_distinct show "finite M" by auto
  qed
qed

lemma lookup_filter: assumes f: "f x"
  shows "lookup x [(x',y) \<leftarrow> xys. f x'] = lookup x xys"
proof (induct xys)
  case Nil show ?case by simp
next
  case (Cons xy xys)
  obtain x' y where xy: "xy = (x',y)" by force
  show ?case
    by (cases "x' = x", insert f xy Cons, auto)
qed

lemma infinite_inj_on_remove_one: fixes f :: "'a \<Rightarrow> 'b" 
  assumes inf: "infinite B"
  and inj: "inj_on f A"
  and ran: "range f \<subseteq> B"
  shows "\<exists> g. inj_on g A \<and> range g \<subseteq> B - {b}"
proof -
  from inf have inf: "infinite (B - {b})" by auto
  from infinite_imp_elem[OF inf] obtain b' where b': "b' \<in> B - {b}" by force
  from infinite_countable_subset[OF inf] obtain h :: "nat \<Rightarrow> 'b" where inj_h: "inj h" and ran_h: "range h \<subseteq> B - {b}" by auto
  let ?h = "\<lambda> a. h (Suc (the_inv h (f a)))"
  let ?f = "\<lambda> a. (if f a \<in> range h then ?h a else f a)"
  let ?g = "\<lambda> a. if f a = b then h 0 else if a \<in> A then ?f a else b'"
  show ?thesis
  proof (rule exI[of _ ?g], unfold inj_on_def, intro conjI ballI impI)
    fix a1 a2
    assume a1: "a1 \<in> A" and a2: "a2 \<in> A" and id: "?g a1 = ?g a2"
    show "a1 = a2" 
    proof (cases "f a1 = b")
      case True 
      with id have id: "?g a2 = h 0" and fa1: "f a1 = b" by auto
      show ?thesis
      proof (cases "f a2 = b")
        case True
        with fa1 id a1 a2 inj show ?thesis unfolding inj_on_def by auto
      next
        case False
        with id a2 have id: "?f a2 = h 0" by auto
        with inj_h show ?thesis unfolding inj_on_def by (cases "f a2 \<in> range h", auto)
      qed
    next
      case False
      with id a1 have id: "?f a1 = ?g a2" by auto
      show ?thesis
      proof (cases "f a2 = b")
        case True
        with id have "?f a1 = h 0" by auto
        with inj_h show ?thesis unfolding inj_on_def by (cases "f a1 \<in> range h", auto)
      next
        case False
        with id a2 have id: "?f a1 = ?f a2" by auto
        have "f a1 = f a2"
        proof (cases "f a1 \<in> range h")
          case True note r1 = this
          with id have id: "?f a2 = ?h a1" by simp
          hence r2: "f a2 \<in> range h" by (cases "f a2 \<in> range h", auto)
          with id have id: "?h a1 = ?h a2" by simp
          from r1 r2 obtain i1 i2 where h: "f a1 = h i1" "f a2 = h i2" by auto
          from id[unfolded h the_inv_f_f[OF inj_h]] have "h (Suc i1) = h (Suc i2)" .
          with inj_h have id: "i1 = i2" unfolding inj_on_def by auto
          with h show "f a1 = f a2" by auto
        next
          case False note r1 = this
          {
            assume "f a2 \<in> range h"
            with id have id: "?f a1 = ?h a2" by simp
            hence "f a1 \<in> range h" by (cases "f a1 \<in> range h", auto)
            with r1 have False by auto
          }
          with r1 id show ?thesis by auto
        qed
        with a1 a2 inj show ?thesis unfolding inj_on_def by auto
      qed
    qed
  next
    show "range ?g \<subseteq> B - {b}" using ran_h ran b' by auto
  qed
qed  

lemma infinite_inj_on_finite_remove_finite: fixes f :: "'a \<Rightarrow> 'b" and B :: "'b set"
  assumes inf: "infinite B"
  and inj: "inj_on f A"
  and fin: "finite B'"
  and ran: "range f \<subseteq> B"
  shows "\<exists> g. inj_on g A \<and> range g \<subseteq> B - B'"
  using fin
proof (induct B')
  case empty
  show ?case using inj ran by auto
next
  case (insert b B')
  from insert(3) obtain g where inj: "inj_on (g :: 'a \<Rightarrow> 'b) A" and ran: "range g \<subseteq> B - B'" by auto
  from insert(1) inf have "infinite (B - B')" by auto
  from infinite_inj_on_remove_one[OF this inj ran, of b]
  show ?case by auto
qed

(* TODO remove in next release *)
lemma floor_lesseq_add: "floor (x :: 'a :: floor_ceiling) + floor y \<le> floor (x + y)"
  by (metis add_le_cancel_left floor_add_of_int floor_mono of_int_floor_le)

lemma ceiling_lesseq_add: "ceiling (x + y) \<le> ceiling (x :: 'a :: floor_ceiling) + ceiling y" using floor_lesseq_add[of "- x" "- y"] unfolding floor_minus
  using floor_minus[of "x + y"] by simp

end
